from __future__ import annotations
import abc
import shapely
from shapely.geometry.base import BaseGeometry
from ..interface import ShapeInterface

MIN_AREA = 1e-7


class Shape(ShapeInterface['Shape'], abc.ABC):
    """
    平面图形计算库的 shapely - impl，接口定义参考 Interface，类型结构如下所示：

    Shape:      一切类型的超类，标识符代称
        EMPTY:      唯一实例，代表空集，表示平面上没有任何元素在 Shape 中
        FULL:       唯一实例，代表全集，表示全平面都在 Shape 中
        Single:         若 Shape 中的全部元素相互联通，称之为”单一图形“，对应”多组图形“ —— 但 EMPTY 和 FULL 不属于 Single
            Region:                 矩形
            Circle：                 圆形
            SimplePolygon:          简单多边形，”简单“在这里指的是”单连通“，对应”复连通“
            ComplexPolygon:         复杂多边形，”复杂“在这里指的是”复连通“，对应”单连通“
        Multi:         若 Shape 中的至少存在两个元素相互不能联通，称之为”多组图形“，对应”单一图形“ —— 但 EMPTY 和 FULL 不属于 Multi
            MultiSimplePolygon:      多组简单多边形
            MultiComplexPolygon:     多组复杂多边形

    基于本人需求考虑，全部图形均用轮廓数组表示，包括圆，也直接写作一族点列，后续不再区分”图形“与”轮廓“的概念
    """
    EMPTY = None
    FULL = None

    __slots__ = ('_geo', '_reversed')

    def __bool__(self) -> bool:
        return self.geo is not None and not self.geo.is_empty

    def __str__(self) -> str:
        tp = type(self).__name__
        rvs = "neg" if self.reversed else "pos"
        xy = tuple(round(p, 1) for p in self.center)
        area = round(self.area, 2)
        bounds = tuple(round(p, 1) for p in self.bounds)
        return f'JassorShape:{tp} {rvs} at {xy} with area {area} in region {bounds}'

    @property
    def geo(self) -> BaseGeometry:
        return self._geo

    @property
    def reversed(self) -> bool:
        return self._reversed

    def is_valid(self) -> bool:
        geo = self.geo
        return geo is not None and geo.is_valid and not geo.is_empty and geo.area > 0

    # @abc.abstractmethod
    # def clean(self):
    #     raise NotImplementedError

    @staticmethod
    def map_cls(tp: str) -> type:
        return {
            'REGION': Single.REGION,
            'CIRCLE': Single.CIRCLE,
            'SIMPLEPOLYGON': Single.SIMPLE,
            'COMPLEXPOLYGON': Single.COMPLEX,
            'MULTISIMPLEPOLYGON': Multi.SIMPLE,
            'MULTICOMPLEXPOLYGON': Multi.COMPLEX,
        }[tp]


class Single(Shape, abc.ABC):
    # 单形,复形的定义
    REGION = None
    CIRCLE = None
    SIMPLE = None
    COMPLEX = None

    __slots__ = ()

    @staticmethod
    def asSimple(shape: Shape) -> Single:
        shape = Single.asComplex(shape)
        return Single.SIMPLE(single=shape.outer, reverse=shape.reversed)

    @staticmethod
    def asComplex(shape: Shape) -> Single:
        if not shape:
            return Single.COMPLEX(reverse=shape.reversed)
        if isinstance(shape, Single):
            geo = shape.geo
        else:
            # 多轮廓是可以兼容单轮廓的，所以也应该提供相应的向回转换方法
            geo = shape.geo
            assert isinstance(geo, shapely.MultiPolygon), f'geo 类型有问题: {geo}'
            if len(geo.geoms) == 0:
                # 没有轮廓就直接给个空的就行
                geo = shapely.Polygon()
            elif len(geo.geoms) == 1:
                # 只有一个轮廓那最好
                geo = geo.geoms[0]
            elif ConvertMulti2SingleException.mode == 'ignore':
                # 忽略模式
                geos = list(geo.geoms)
                geos.sort(key=lambda g: g.area, reverse=True)
                geo = geos[0]
            elif ConvertMulti2SingleException.mode == 'smart':
                # 智慧模式
                geos = list(geo.geoms)
                geos.sort(key=lambda g: g.area, reverse=True)
                if abs(1 - geos[0].area / geo.area) < ConvertMulti2SingleException.thresh:
                    geo = geos[0]
                else:
                    raise ConvertMulti2SingleException(f'Multi 轮廓长度为 {len(geo.geoms)}，'
                                                       f'其中最大的单一轮廓面积是 {geos[0].area}，总面积是 {geo.area}，'
                                                       f'占比为 {geos[0].area / geo.area} 低于{1 - ConvertMulti2SingleException.thresh}，'
                                                       f'不能转换')
            else:
                # 报错模式
                raise ConvertMulti2SingleException(f'Multi 轮廓长度为 {len(geo.geoms)}，不能转换')
        return Single.COMPLEX(geo=geo, reverse=shape.reversed)


class Multi(Shape, abc.ABC):
    # 多凸形,多单形,多复形的定义
    CONVEX = None
    SIMPLE = None
    COMPLEX = None

    __slots__ = ()

    @staticmethod
    def asSimple(shape: Shape) -> Multi:
        shape = Multi.asComplex(shape)
        return Multi.SIMPLE(geo=shape.outer.geo, reverse=shape.reversed)

    @staticmethod
    def asComplex(shape: Shape) -> Multi:
        if not shape:
            return Multi.COMPLEX(reverse=shape.reversed)
        if isinstance(shape, Multi):
            geo = shape.geo
        else:
            # 多轮廓可以兼容单轮廓，但必须要改类型，因为只有多轮廓里才有 .geoms 属性
            geo = shapely.MultiPolygon(polygons=[shape.geo])
        return Multi.COMPLEX(geo=geo, reverse=shape.reversed)

    def __len__(self) -> int:
        return len(self.geo.geoms)


class ConvertMulti2SingleException(Exception):
    mode = 'error'
    thresh = 0.05
    """
    当此处 mode 设置为 error 时，一个 MultiShape 转化为 SingleShape 会报错
    当此处 mode 设置为 ignore 时，一个 MultiShape 转化为 SingleShape 时会直接选择面积最大的作为结果
    当此处 mode 设置为 smart 时，会根据最大面积占总面积比例决定报错还是忽略
    """

    @staticmethod
    def mode_error():
        ConvertMulti2SingleException.mode = 'error'

    @staticmethod
    def mode_ignore():
        ConvertMulti2SingleException.mode = 'ignore'

    @staticmethod
    def mode_smart(thresh: float = 0.05):
        ConvertMulti2SingleException.mode = 'smart'
        ConvertMulti2SingleException.thresh = thresh


class CoordinatesNotLegalException(Exception):
    pass


class NoParametersException(Exception):
    pass



import json
import pickle
from typing import Tuple, Union, Optional, IO, List
import shapely
from shapely.geometry.base import BaseGeometry

from .definition import Shape, Single, Multi, MIN_AREA


"""
这一批函数之所以单独放到 functional 里，是因为它们的逻辑调用关系与 reverse 高度相关
如果不拆成两族函数，代码写起来会非常臃肿
所以你会发现，这里只有一些集合论相关的运算
并且这里是直接对接 shapely 库的实现，并不处理基于 reverse 的逻辑关系（这些关系在上一层函数中已经处理过了）
"""


Position = Union[str, complex, Tuple[float, float]]


def inter(*shapes: Shape, reverse: bool) -> Shape:
    geo = None
    for shape in shapes:
        if geo is None: geo = shape.geo
        else: geo = geo.intersection(shape.geo)
    return norm_multi(geo, reverse=reverse)


def union(*shapes: Shape, reverse: bool) -> Shape:
    geo = None
    for shape in shapes:
        if geo is None: geo = shape.geo
        else: geo = geo.union(shape.geo)
    return norm_multi(geo, reverse=reverse)


def diff(self: Shape, other: Shape, reverse: bool) -> Shape:
    geo = self.geo.symmetric_difference(other.geo)
    return norm_multi(geo, reverse=reverse)


def remove(self: Shape, other: Shape, reverse: bool) -> Shape:
    geo = self.geo.difference(other.geo)
    return norm_multi(geo, reverse=reverse)


def norm_single(geo: BaseGeometry, reverse: bool) -> Single:
    geo = norm_geo(geo)
    if geo is None:
        return Shape.FULL if reverse else Shape.EMPTY
    if isinstance(geo, shapely.Polygon):
        return Single.COMPLEX(geo=geo)
    # 这里如果 geo 跑出来是 multi，而我需要将它按 single 去标准化，就有可能出现异常，异常控制代码在 Single.asComplex 方法里
    multi = Multi.COMPLEX(geo=geo, reverse=reverse)
    return Single.asComplex(multi)


def norm_multi(geo: BaseGeometry, reverse: bool) -> Multi:
    geo = norm_geo(geo)
    if geo is None:
        return Shape.FULL if reverse else Shape.EMPTY
    # 这里如果 geo 跑出来是 multi，而我需要把它变成 single，就需要调用 as 方法了
    if isinstance(geo, shapely.Polygon):
        geo = shapely.MultiPolygon(polygons=[geo])
    return Multi.COMPLEX(geo=geo, reverse=reverse)


def norm_geo(geo: BaseGeometry) -> Optional[BaseGeometry]:
    # 输入检查，处理各种妖魔鬼怪，这里比较重要的是最小面积阈值，需要在每次使用时定义
    if geo is None or geo.is_empty or geo.area <= MIN_AREA:
        return None
    # geo 可能存在自相交的曲线, 也可能 multi_geo = [g1, g2] 并且 g1、g2 相交
    # buff(0) 好像可以消灭一切牛鬼蛇神, unary_union 则只能解决后者，遇到前者直接报错
    if not geo.is_valid or not geo.is_simple:
        geo = geo.buffer(0)

    # shapely 中只有 polygon 有面积，因此本工具库也只处理 polygon 或可能包含 polygon 的元素
    if isinstance(geo, shapely.Polygon):
        return geo

    if isinstance(geo, shapely.MultiPolygon):
        return geo

    if isinstance(geo, shapely.GeometryCollection):
        polygons = [g for g in geo.geoms if isinstance(g, shapely.Polygon)]
        geo = shapely.MultiPolygon(polygons=polygons)
        return geo

    return None


def load(f: IO) -> Shape:
    return loads(f.readlines())


def loads(lines: List[str]) -> Shape:
    tp = lines[0].strip().upper()
    if tp == 'EMPTY':
        return Shape.EMPTY
    if tp == 'FULL':
        return Shape.FULL
    tp = Shape.map_cls(tp)
    rvs = bool(lines[1])
    points = json.loads(lines[2])
    return tp(from_p=points, reverse=rvs)


def loadb(f: IO) -> Shape:
    tp, rvs, points = pickle.load(f)
    tp = tp.upper()
    if tp == 'EMPTY':
        return Shape.EMPTY
    if tp == 'FULL':
        return Shape.FULL
    tp = Shape.map_cls(tp)
    return tp(from_p=points, reverse=rvs)



from typing import List, Tuple, Iterable
import shapely
from shapely.geometry.base import BaseGeometry

from .definition import Shape, Single, Multi, NoParametersException, CoordinatesNotLegalException
from .impl_multi_complex import MultiComplexPolygon
from . import functional as F


class MultiSimplePolygon(MultiComplexPolygon):
    """
    多-单连通多边形, 创建方式有四:
    1. 指定 geo
    2. 指定 Multi
    3. 指定 Single 数组
    4. 指定 outers
    遵循逆序优先规则
    """

    __slots__ = ()

    def __init__(
            self,
            outers: List[Tuple[float, float]] = None,
            geo: BaseGeometry = None,
            shapes: Iterable[Shape] = None,
            from_p: List[Tuple[float, float]] = None,
            reverse: bool = False,
    ):
        if geo is not None:
            assert isinstance(geo, shapely.MultiPolygon), 'geo 必须是 MultiPolygon'
            assert all(g.boundary.geom_type.upper() == 'LINESTRING' for g in geo.geoms), 'geo 必须是单连通的'
        elif shapes is not None:
            assert all(isinstance(shape, (Single.SIMPLE, Multi.SIMPLE)) for shape in shapes if shape), 'shapes 必须由 SIMPLE（单连通） 图像构成'
        else:
            if from_p is not None:
                outers = from_p
            if outers is None:
                # 没有任何参数的话，要报个错
                raise NoParametersException(f'Any of such parameters have to be provided: (outer, *inners), geo, single, from_p')

            # 对用户输入进行检查和修复
            geo = shapely.MultiPolygon(polygons=[(outer, []) for outer in outers])
            geo = F.norm_geo(geo)
            # 创建时要求轮廓必须合法
            if geo is None:
                raise CoordinatesNotLegalException(f'creating single polygon with geo=={type(geo)}')
            elif isinstance(geo, shapely.Polygon):
                geo = shapely.MultiPolygon(polygons=[geo])
            # 还得是单连通的
            if any(g.interiors for g in geo.geoms):
                raise CoordinatesNotLegalException(f'MultiSimplePolygon not allowed interiors coordinates with geo interiors nums: {list(len(g.interiors) for g in geo.geoms)}')

        super().__init__(outers=None, geo=geo, shapes=shapes, reverse=reverse)

    # def merge(self, shape: Shape) -> Multi:
    #     # 合集运算
    #     geo = self.geo
    #     singles = shape.sep_out()
    #     geos = [s.geo for s in singles if not geo.disjoint(s.geo)]
    #     for g in geos:
    #         geo = geo.union(g)
    #     geo = self.__norm_multi__(geo)
    #     return ComplexMultiPolygon(geo=geo)

    @property
    def outer(self) -> Multi:
        # 外轮廓(正形)
        return Multi.SIMPLE(geo=self.geo)

    @property
    def inner(self) -> Shape.EMPTY:
        # 内轮廓(负形)
        return Shape.EMPTY

    def sep_in(self) -> Tuple[Multi, Shape.EMPTY]:
        # 内分解
        # 逐层分解 (规避由 shapely 的任意性引起的荒诞错误)
        return self, Shape.EMPTY

    def sep_out(self) -> List[Single]:
        # 外分解
        singles = [Single.SIMPLE(geo=g) for g in self.geo.geoms if isinstance(g, shapely.Polygon)]
        singles = [s for s in singles if s.is_valid()]
        return singles

    def sep_p(self) -> List[List[Tuple[int, int]]]:
        # 点分解
        # 逐层分解 (规避由 shapely 任意性引起的荒诞错误)
        singles = [Single.SIMPLE(geo=g) for g in self.geo.geoms if isinstance(g, shapely.Polygon)]
        singles = [s for s in singles if s.is_valid()]
        return [s.sep_p() for s in singles if s.is_valid()]

    @property
    def cls(self) -> type:
        return MultiSimplePolygon


Multi.SIMPLE = MultiSimplePolygon



from typing import List, Tuple

import math
from shapely.geometry.base import BaseGeometry

from .definition import Single
from .impl_single_simple import SimplePolygon


class Circle(SimplePolygon):
    """
    单-圆形，仅为了方便区域的创建，操作与凸形状完全一致
    实际存储时用的也是多边形，并不会真正创建一个圆
    """
    
    __slots__ = ()

    def __init__(
            self, x: float = 0, y: float = 0, r: float = 0, num: int = 100,
            geo: BaseGeometry = None,
            from_p: List[Tuple[int, int]] = None,
            reverse: bool = False,
    ):
        if geo is not None or from_p is not None:
            outer = None
        else:
            base_directs = [2 * math.pi / num * i for i in range(num)]
            base_points = [(math.cos(d), math.sin(d)) for d in base_directs]
            outer = [(x + p * r, y + q * r) for p, q in base_points]
        super().__init__(outer, geo=geo, from_p=from_p, reverse=reverse)

    @property
    def cls(self) -> type:
        return Circle


Single.CIRCLE = Circle



from typing import Tuple

from shapely.geometry.base import BaseGeometry

from .definition import Single
from .impl_single_simple import SimplePolygon


class Region(SimplePolygon):
    """
    单-矩形，仅为了方便区域的创建，操作与凸形状完全一致
    仅作为类型标识符存在
    """

    __slots__ = ()

    def __init__(
            self, left: float = 0, up: float = 0, right: float = 0, down: float = 0,
            geo: BaseGeometry = None,
            from_p: Tuple[float, float, float, float] = None,
            reverse: bool = False,
    ):
        if geo is not None:
            outer = None
        else:
            if from_p is not None:
                left, up, right, down = from_p
            p1 = (left, up)
            p2 = (left, down)
            p3 = (right, down)
            p4 = (right, up)
            outer = [p1, p2, p3, p4]
        super().__init__(outer, geo=geo, from_p=None, reverse=reverse)

    def sep_p(self) -> Tuple[float, float, float, float]:
        return self.geo.bounds

    @property
    def cls(self) -> type:
        return Region


Single.REGION = Region
